"""Command line formatting, prompts, and other utilities."""

import re
from blessings import Terminal

# Globals

_term = Terminal()

# Keys into COLORS
ERROR = 'error'
ERROR_TITLE = 'error_title'
WARNING = 'warning'
SUCCESS = 'success'
INFO = 'info'
PROMPT = 'prompt'
#: Dictionary mapping indices to formatting functions
COLORS = {
    None: str,
    ERROR: _term.red,
    ERROR_TITLE: _term.bold_red,
    WARNING: _term.yellow,
    SUCCESS: _term.green,
    INFO: _term.cyan,
    PROMPT: _term.magenta,
}
#: String to use when indenting output
INDENT = ' ' * 4


# Output Formatting

def print_multiline(first_line, *lines,
                    formatting=None, first_line_formatting=None,
                    indent=True, indent_first_line=False):
    """Print multiple lines

    :param first_line: First line to print
    :param lines: Positional parameters will each be printed on their own line
    :param formatting: (Optional) Set to a formatting constant to format each
        line
    :param first_line_formatting: (Optional) Set to override formatting for the
        first line
    :param indent: (Default: True) If True and there's multiple lines, indent
        all lines after the first
    :param indent_first_line: (Default: False) If True, indent the first line
    """
    if formatting not in COLORS:
        formatting = None
    if first_line_formatting is None or first_line_formatting not in COLORS:
        first_line_formatting = formatting
    fmt_func = COLORS[formatting]
    first_line_fmt_func = COLORS[first_line_formatting]

    first_line_prefix = INDENT if indent_first_line else ''
    print(first_line_fmt_func(first_line_prefix + str(first_line)))

    if lines:
        line_prefix = INDENT if indent else ''
        for line in lines:
            print(fmt_func(line_prefix + str(line)))


def print_error(*lines):
    print_multiline(*lines, first_line_formatting=ERROR_TITLE, formatting=ERROR)


def print_warning(*lines):
    print_multiline(*lines, formatting=WARNING)


def print_success(*lines):
    print_multiline(*lines, formatting=SUCCESS)


def print_info(*lines):
    print_multiline(*lines, formatting=INFO)


# User Input Prompts

def sanitize_input(val):
    return val.strip()


class ValidationError(Exception):
    """Raised if input validation fails"""
    pass


def validate_optional_prompt(val, error_msg=None):
    """Dummy validation function for optional prompts. Just returns val"""
    # TODO Should there just be an option in prompt()? If input is non-blank you'll probably still wanna validate it
    return val


def validate_nonempty(val, error_msg=None):
    """Raises ValidationError if val is empty

    :param val: Input to validate
    :param error_msg: (Optional) Custom error text to print if val is invalid

    :return: Validated input
    """
    if not val:
        raise ValidationError(error_msg or 'Please enter some text.')
    return val


def validate_yn(val, error_msg=None):
    """Validate y/n prompts

    :param val: User response to y/n prompt. If a boolean value is passed
        (e.g. if a prompt received initial_input=True), it is treated as a y/n
        answer and considered valid input
    :param error_msg: (Optional) Custom error text to print if val is invalid

    :return: True if user answered yes, False if user answered no
    """
    # If a boolean value was passed, return it
    if isinstance(val, bool):
        return val
    val = val.lower().strip()
    if val not in ['y', 'yes', 'n', 'no']:
        raise ValidationError(error_msg or 'Please enter "y" or "n".')
    return val in ['y', 'yes']


def generate_validate_regex_function(expr, default_error_msg='No matches found.', show_expr_in_error_msg=True):
    """Generate a validation function that validates a given regular expression.

    :param expr: Regular expression to validate against
    :param default_error_msg: (Optional) Default validation error message to use
        in this validation function
    :param show_expr_in_error_msg: (Default: True) If True, expr will be shown
        in error message

    :return: Generated validation function
    """
    def validate_regex(val, error_msg=None):
        if not error_msg:
            error_msg = default_error_msg
        if show_expr_in_error_msg:
            error_msg += '\n' + INDENT + 'Must match regex: ' + expr
        res = re.findall(expr, val)
        if not res:
            raise ValidationError(error_msg)
        return res[0]
    return validate_regex


def prompt(prompt_text, *extended_description,
           initial_input=None, default_val=None,
           sanitize_function=sanitize_input, validate_function=validate_nonempty, format_function=None,
           invalid_msg=None, print_newline_on_success=True):
    """Prompt user for input

    :param prompt_text: Text to display next to input area
    :param *extended_description: (Optional) Explanation of prompt. Printed
        before prompt. Each positional parameter here is printed on its own
        line

    :param initial_input: (Optional) Initial input (e.g. something passed in
        via command line args). If it passes validation, will use this and skip
        the input prompt
    :param default_val: (Optional) Value to use if no input is provided

    :param sanitize_function: (Default: sanitize_input) Function to sanitize
        input
    :param validate_function: (Default: validate_nonempty) Function used to
        validate input
    :param format_function: (Optional) Function to format input after running
        it through sanitize_function and before passing it to validate_function

    :param invalid_msg: (Optional) Error message text to display if validation
        fails
    :param print_newline_on_success: (Default: True) Print a newline after
        successfully getting valid input

    :return: Input after sanitization, formatting, and validation
    """
    # If input for this prompt was given via an argument, attempt to validate
    # it and bypass prompt
    if initial_input is not None:
        try:
            initial_input = sanitize_function(initial_input)
            if format_function is not None:
                initial_input = format_function(initial_input)
            val = validate_function(initial_input, invalid_msg)
        except ValidationError as e:
            print_error(e)
        else:
            return val
    # Print description and prompt
    if extended_description:
        print(*extended_description, sep='\n')
    text = COLORS[PROMPT]('> ' + (
        f'{prompt_text} [{default_val}]'
        if default_val is not None else
        prompt_text
    ) + ': ')
    # Loop until we get valid input
    while True:
        val = sanitize_function(input(text))
        # If we have a default val, use it if input is empty
        if not val and default_val is not None:
            val = default_val
        # Format val
        if format_function is not None:
            val = format_function(val)
        # Attempt to validate, loop if invalid
        try:
            val = validate_function(val, invalid_msg)
        except ValidationError as e:
            print_error(e)
            continue
        break
    if print_newline_on_success:
        print('')
    return val




"""File-related utilities"""
import re


def sanitize_filename(filename, allow_spaces=False):
    """Strips invalid characters from a filename.

    Considers `POSIX "fully portable filenames"
    <http://pubs.opengroup.org/onlinepubs/9699919799/basedefs/V1_chap03.html#tag_03_282>`__
    valid. These include:

        A-Z a-z 0-9 ._-

    Filenames cannot begin with a hyphen.

    :param filename: The desired file name (without path)
    :param allow_spaces: (Default = False) If True, spaces will be considered
        valid characters

    :return: Filename with invalid characters removed
    """
    regex = r'^-|[^\d\w\. -]' if allow_spaces else r'^-|[^\d\w\.-]'
    return re.sub(regex, '', filename)



from git import GitCommandError
from git_workflow.utils import cmd
from git_workflow.utils.repository import checkout_branch
from .base import WorkflowBase
from .unset_template import UnsetTemplate


class FinishBranch(WorkflowBase):
    """\
    Finish a project branch.

    By default, this command will prompt for confirmation unless ``--force`` is
    specified. Once confirmed, this command will:

    - Unset the commit template of the project branch
    - Checkout the base branch and pull latest updates
    - Attempt to delete the project branch using ``git branch -d``, which may
      fail if the project branch has not been fully merged
    """

    command = 'finish'
    description = 'Finish a project branch.'
    configs_used = ['baseBranch', 'finishBranchConfirmationPrompt']

    @classmethod
    def add_subparser(cls, subparsers, generic_parent_parser):
        finish_subparser = cls._add_base_subparser(subparsers, generic_parent_parser)
        # TODO: no_pull? -D?
        # Specify branch
        positional_args = finish_subparser.add_argument_group(
            'Positional Arguments'
        )
        positional_args.add_argument(
            'branch', metavar='<branch>', nargs='?',
            help='Branch to finish (default: current)',
            default=None
        )
        # Confirmation prompt
        confirmation_args = finish_subparser.add_argument_group(
            'Confirmation Prompt Arguments',
            'Override workflow.finishBranchConfirmationPrompt config.'
        )
        confirmation_group = confirmation_args.add_mutually_exclusive_group()
        confirmation_group.add_argument(
            '-f', '--force', help='Skip confirmation prompt (if configured)',
            dest='confirm', action='store_false', default=None
        )
        confirmation_group.add_argument(
            '-c', '--confirmation', help='Prompt for confirmation before deleting',
            dest='confirm', action='store_true', default=None
        )

    def get_args(self):
        args = {}
        args['branch'] = (self.parsed_args.branch
                          if self.parsed_args.branch is not None else
                          self.repo.active_branch.name)
        # Default to config value unless otherwise specified
        args['confirm'] = (self.configs.FINISH_BRANCH_CONFIRMATION_PROMPT
                           if self.parsed_args.confirm is None else
                           self.parsed_args.confirm)
        return args

    def run(self):
        args = self.get_args()
        branch = args['branch']
        # Confirmation prompt
        if args['confirm']:
            confirmation = cmd.prompt(
                'Delete Branch? (y/n)',
                f'Delete branch {branch}?',
                default_val='n', validate_function=cmd.validate_yn
            )
            if not confirmation:
                return
        # Unset template
        unset_template_parsed_args = self.parser.parse_args([UnsetTemplate.command, branch, '--force'])
        unset_template = UnsetTemplate(self.repo, self.parser,
                                       parsed_args=unset_template_parsed_args,
                                       verbosity=self.verbosity)
        unset_template.run()
        # Checkout base_branch
        base_branch = self.configs.BASE_BRANCH
        base_head = checkout_branch(self.repo, base_branch)
        # Finish branch
        self.print(f'Attempting to delete {branch}...')
        try:
            self.repo.git.branch('-d', branch)
            self.print_success('Deletion successful.')
        except GitCommandError:
            self.print_warning(f'Unable to delete {branch}. You will need to delete it manually.')



"""Base class for workflow scripts."""
from abc import ABC, abstractmethod
from git_workflow.utils import cmd
from git_workflow.utils.configs import Configs


class WorkflowBase(ABC):

    def __init__(self, repo, parser,
                 parsed_args=None, verbosity=1):
        """Constructor

        :param repo: git.Repo instance for the repository
        :param parser: ArgumentParser instance
        :param parsed_args: (Optional) Parsed args object
        :param verbosity: (Default: 1) Output verbosity level
        """
        self.repo = repo
        self.parser = parser
        self.parsed_args = parsed_args
        self.verbosity = verbosity
        self.configs = Configs(self.repo)

    #: Names of git configs used in this command
    configs_used = []

    @classmethod
    def _add_base_subparser(cls, subparsers, generic_parent_parser):
        """Common setup for initializing subparser.

        :param subparsers: Subparsers object
        :param generic_parent_parser: Instance of argparse.ArgumentParser used
            as a parent for top level parser

        :return: Subparser object
        """
        return subparsers.add_parser(
            cls.command, description=cls.description, help=cls.description,
            parents=[generic_parent_parser], add_help=False
        )

    # Abstract Properties and Methods

    @property
    @abstractmethod
    def command(self):
        """Subcommand name"""
        pass

    @property
    @abstractmethod
    def description(self):
        """Subcommand description"""
        pass

    @classmethod
    @abstractmethod
    def add_subparser(cls, subparsers, generic_parent_parser):
        """Add subparser for this command to an argument parser

        :param subparsers: Subparsers object
        :param generic_parent_parser: Instance of argparse.ArgumentParser used
            as a parent for top level parser
        """
        pass

    @abstractmethod
    def run(self):
        """Execute the script"""
        pass

    # Helper Methods

    def print(self, *lines, required_verbosity=1, **print_multiline_kwargs):
        """Print a message.

        :param lines: Lines to print
        :param required_verbosity: (Default: 1) Only print if verbosity is greater than
            or equal to this value

        :param **print_multiline_kwargs: Any keyword arguments to pass to
            cmd.print_multiline()
        """
        if self.verbosity >= required_verbosity:
            cmd.print_multiline(*lines, **print_multiline_kwargs)

    def print_error(self, *lines, required_verbosity=1):
        if self.verbosity >= required_verbosity:
            cmd.print_error(*lines)

    def print_warning(self, *lines, required_verbosity=1):
        if self.verbosity >= required_verbosity:
            cmd.print_warning(*lines)

    def print_success(self, *lines, required_verbosity=1):
        if self.verbosity >= required_verbosity:
            cmd.print_success(*lines)

    def print_info(self, *lines, required_verbosity=1):
        if self.verbosity >= required_verbosity:
            cmd.print_info(*lines)



"""Utilities for common interactions with git repo."""
import os
from git import GitCommandError, Head, Remote
from git.cmd import Git as GitCmd
from git_workflow.__about__ import __min_git_version__


# Git Verification + Workflow Config Methods

def verify_git_version(strict=True):
    """Returns True if minimum git version is met for advanced features"""
    g = GitCmd()
    major, minor, patch = g.version_info
    version_float = float(f'{major}.{minor}')
    is_version_requirement_met = version_float >= __min_git_version__
    if strict and not is_version_requirement_met:
        error_msg = f'Git version of {__min_git_version__} or higher required (current version: {version_float})'
        raise Exception(error_msg)
    return is_version_requirement_met


def get_workflow_config_path(repo):
    """Returns the full path for the git workflow config file."""
    return os.path.join(repo.git_dir, 'config_workflow')


def verify_workflow_config_file(repo):
    """Returns True if config_workflow file exists for this repo."""
    return os.path.exists(get_workflow_config_path(repo))


def verify_workflow_config_include(repo):
    """Returns True if config_workflow is included in local git config."""
    result = False
    try:
        included_paths = repo.git.config('include.path', get_all=True, local=True).split('\n')
        result = 'config_workflow' in included_paths
    except GitCommandError:
        pass
    return result


# Initialize

def create_workflow_config_file(repo):
    """Create config_workflow and configure workflow.configpath.

    :return: Result of verify function after attempting to create config
    """
    workflow_config_path = get_workflow_config_path(repo)
    repo.git.config('workflow.configpath', workflow_config_path, file=workflow_config_path)
    return verify_workflow_config_file(repo)


def include_config_workflow(repo):
    """Include config_workflow in repo's local config.

    :return: Result of verify function after attempting to include config
    """
    repo.git.config('include.path', 'config_workflow', add=True, local=True)
    return verify_workflow_config_include(repo)


def initialize(repo):
    """Initialize the repo for use with workflow scripts (unless already
    initialized).

    :return: True if initialized, False if attempted to initialize and
        something went wrong
    """
    workflow_config_file_exists = verify_workflow_config_file(repo)
    workflow_config_included = verify_workflow_config_include(repo)
    if not workflow_config_file_exists:
        print('Workflow config file not found for repo, creating...')
        workflow_config_file_exists = create_workflow_config_file(repo)
        if workflow_config_file_exists:
            print('Config created.')
            print('')
        else:
            pass # TODO ERROR
    if not workflow_config_included:
        print('Configuring local repo to include workflow config...')
        workflow_config_included = include_config_workflow(repo)
        if workflow_config_included:
            print('Repo configured.')
            print('')
        else:
            pass # TODO ERROR
    # TODO don't bother returning once exceptions are implemented
    return workflow_config_file_exists and workflow_config_included


# Common Git Actions

def checkout_branch(repo, branch_name, no_pull=False):
    """Checkout a branch and optionally pull updates.

    :param repo: Repo object
    :param branch_name: Name of the branch to checkout
    :param no_pull: (Default: False) If True, don't pull changes to branch

    :return: Head object for the checked out branch
    """
    base_head = Head(repo, f'refs/heads/{branch_name}')
    if repo.active_branch != base_head:
        print(f'Checking out {branch_name}.')
        base_head.checkout()
    if not no_pull and base_head.tracking_branch():
        print(f'Pulling updates to {branch_name}...')
        remote_name = base_head.tracking_branch().remote_name
        remote = Remote(repo, remote_name)
        base_commit = base_head.commit
        for fetch_info in remote.pull():
            if fetch_info.ref == base_head.tracking_branch():
                if fetch_info.commit != base_commit:
                    print(f'Updated {branch_name} to {fetch_info.commit.hexsha}')
                else:
                    print(f'{branch_name} already up to date.')
        print('')
    return base_head


def fetch_tags(repo):
    """Shorthand for git fetch --all --tags

    :param repo: Repo object
    """
    if repo.remotes:
        print('Fetching tags from remote...')
        print('')
        repo.git.fetch(all=True, tags=True)



# PYTHON_ARGCOMPLETE_OK
import argcomplete
from git_workflow.utils.parser import get_parser
# Initialize parser before remaining imports for improved tab speed
parser = get_parser()
argcomplete.autocomplete(parser)
import os
import sys
from git import Repo
from git.exc import InvalidGitRepositoryError, NoSuchPathError
from git_workflow.utils import cmd, repository
from git_workflow.workflow import run_command


def main():
    # Argument Parser
    parsed_args = parser.parse_args()
    # Check installed git version
    try:
        repository.verify_git_version()
    except Exception as e:
        cmd.print_error(e)
        return
    # Initialize Repo object
    repo = None
    try:
        repo = Repo(os.getcwd(), search_parent_directories=True)
    except InvalidGitRepositoryError as e:
        cmd.print_error('No git repo found: {}'.format(e))
    except NoSuchPathError as e:
        cmd.print_error('Invalid path: {}'.format(e))
    finally:
        if repo is None:
            return
    try:
        run_command(repo, parser, parsed_args=parsed_args)
    except KeyboardInterrupt:
        print('')
        sys.exit(0)
    except Exception as e:
        cmd.print_error(e)
        sys.exit(1)


if __name__ == '__main__':
    main()



import datetime
import re
from git_workflow.utils import cmd
from git_workflow.utils.repository import checkout_branch, fetch_tags
from .base import WorkflowBase
from .set_template import SetTemplate


class StartBranch(WorkflowBase):
    """\
    Create a new branch with the following name format:

    ::

        [<client>-]<brief-description>-<yyyymmdd>-<initials>

    Where:

    - ``<client>`` - (Optional) Client's name
    - ``<brief-description>`` - Description of the work
    - ``<yyyymmdd>`` - Today's date
    - ``<initials>`` - Engineer's initials

    Script will prompt for details and format appropriately (i.e. no
    spaces/underscores, all lowercase).
    """

    command = 'start'
    description = 'Create a new branch.'
    configs_used = ['initials', 'baseBranch', 'badBranchNamePatterns']

    @classmethod
    def add_subparser(cls, subparsers, generic_parent_parser):
        branch_subparser = cls._add_base_subparser(subparsers, generic_parent_parser)
        # Branch Name
        branch_name_args = branch_subparser.add_argument_group(
            'Branch Name Arguments'
        )
        client_group = branch_name_args.add_mutually_exclusive_group()
        client_group.add_argument(
            '-c', '--client', metavar='<client>', help='Specify client name'
        )
        client_group.add_argument(
            '-C', '--no-client', help='No client name (skips prompt)',
            action='store_true', default=False
        )
        branch_name_args.add_argument(
            '-d', '--description', metavar='<description>', help='Specify branch description'
        )
        branch_name_args.add_argument(
            '-i', '--initials', metavar='<initials>', help='Specify developer initials'
        )
        branch_name_args.add_argument(
            '-s', '--skip-bad-name-check', help='Skip check for bad branch names',
            action='store_true', default=False
        )
        # Commit Template
        commit_template_args = branch_subparser.add_argument_group(
            'Commit Template Arguments'
        )
        ticket_group = commit_template_args.add_mutually_exclusive_group()
        ticket_group.add_argument(
            '-t', '--ticket', metavar='<ticket#>', help='Specify ticket number (will create commit template)'
        )
        ticket_group.add_argument(
            '-T', '--no-ticket', help="Skip ticket number prompt, don't create commit template (overrides -t)",
            action='store_true', default=False
        ) # https://youtu.be/iHSPf6x1Fdo
        # Branching
        branching_args = branch_subparser.add_argument_group(
            'Branching Arguments'
        )
        base_branch_group = branching_args.add_mutually_exclusive_group()
        base_branch_group.add_argument(
            '-b', '--base-branch', metavar='<branch>', help='Specify branch to use as base for new branch (default: master)'
        )
        base_branch_group.add_argument(
            '-B', '--branch-from-current', help='Use currently checked out branch as base (overrides -b)',
            action='store_true', default=False
        )
        base_branch_group.add_argument(
            '-r', '--base-release', metavar='<tag>', help='Branch from the specified git tag'
        )
        branching_args.add_argument(
            '-P', '--no-pull', help='Skip pulling changes to base branch',
            action='store_true', default=False
        )

    def get_args(self):
        """Parse command line arguments and prompt for any missing values.

        :return: A dictionary with the following keys:
            client, description, initials, ticket, timestamp, base_branch,
            skip_bad_name_check
        """
        args = {}
        client = None
        if not self.parsed_args.no_client:
            client = cmd.prompt(
                'Client',
                '(Optional) Enter the name of the affected client.',
                initial_input=self.parsed_args.client,
                validate_function=cmd.validate_optional_prompt,
                format_function=self.format_branch_name,
            )
        # Append hyphen if client is not empty
        if client:
            client += '-'
        else:
            client = ''
        args['client'] = client

        description = cmd.prompt(
            'Description',
            'Enter a brief description for the branch.',
            invalid_msg='Description must not be blank.',
            initial_input=self.parsed_args.description,
            format_function=self.format_branch_name,
        )
        description += '-'
        args['description'] = description

        initials = cmd.prompt(
            'Initials',
            'Enter your initials.',
            invalid_msg='Must enter initials.',
            initial_input=self.parsed_args.initials or self.configs.INITIALS,
            format_function=self.format_branch_name,
        )
        args['initials'] = initials

        ticket = None
        if not self.parsed_args.no_ticket:
            ticket = cmd.prompt(
                'Ticket Number',
                '(Optional) Enter ticket number to use in commit messages.',
                "Leave blank if you don't want to use a commit template.",
                initial_input=self.parsed_args.ticket,
                validate_function=cmd.validate_optional_prompt,
            )
        args['ticket'] = ticket

        timestamp = datetime.datetime.now().strftime('%Y%m%d-')
        args['timestamp'] = timestamp

        base_branch = self.parsed_args.base_branch or self.configs.BASE_BRANCH
        if self.parsed_args.branch_from_current:
            base_branch = self.repo.active_branch.name
        args['base_branch'] = base_branch

        args['base_release'] = self.parsed_args.base_release

        args['no_pull'] = self.parsed_args.no_pull

        args['skip_bad_name_check'] = self.parsed_args.skip_bad_name_check

        return args

    def run(self):
        args = self.get_args()
        branch_name = args['client'] + args['description'] + args['timestamp'] + args['initials']
        if self.configs.BAD_BRANCH_NAME_PATTERNS:
            if args['skip_bad_name_check']:
                self.print_warning('workflow.badBranchNamePatterns is configured, but -s argument was specified.',
                                   'Skipping bad branch name check.')
            else:
                # Will raise exception if name doesn't check out
                self.check_branch_name(branch_name)
        # Checkout base branch or tag
        base_branch = args['base_branch']
        base_release = args['base_release']
        new_active_branch = None
        # base_release will only be set if the --base-release arg is specified, overrides base branch
        if base_release is None:
            base_head = checkout_branch(self.repo, base_branch, no_pull=args['no_pull'])
            # Checkout new branch
            self.print(f'Creating new branch {branch_name}...')
            new_active_branch = base_head.checkout(b=branch_name)
        else:
            # Update
            if not args['no_pull']:
                fetch_tags(self.repo)
            self.print(f'Creating new branch {branch_name} based on tag {base_release}...')
            self.repo.git.checkout(base_release, b=branch_name)
            new_active_branch = self.repo.active_branch
        # Verify branch
        if new_active_branch is not None and new_active_branch.name == branch_name:
            self.print_success('Branch created.', '')
        else:
            # Not sure we'd ever get here, but print an error anyway to be safe
            raise Exception(f'Unable to create branch {branch_name} due to an unknown error.')
        # If specified, call commit-template
        if args['ticket']:
            self.print('Checking ticket number format...')
            set_template_parsed_args = self.parser.parse_args([SetTemplate.command, args['ticket']])
            set_template = SetTemplate(self.repo, self.parser,
                                       parsed_args=set_template_parsed_args,
                                       verbosity=self.verbosity)
            set_template.run()

    # Helper Methods

    def format_branch_name(self, val):
        """Convert text to lowercase and replace spaces and underscores with
        hyphens.

        :param val: Text to format

        :return: Formatted text
        """
        return re.sub('[ _]+', '-', val.lower())

    def check_branch_name(self, branch_name):
        """Checks for configured bad patterns in branch name. Raises exception
        if one is found.

        :param branch_name: Name to check
        """
        for bad_pattern in self.configs.BAD_BRANCH_NAME_PATTERNS:
            if bad_pattern in branch_name:
                error_msg = '\n'.join([
                    f'Branch name "{branch_name}" contains invalid pattern "{bad_pattern}".',
                    '',
                    'Branch names should not include the following patterns:',
                    *[cmd.INDENT + pattern for pattern in self.configs.BAD_BRANCH_NAME_PATTERNS],
                    '',
                    '(from git config workflow.badBranchNamePatterns)',
                    '',
                    'To skip this check, use --skip-bad-name-check argument.'
                ])
                raise Exception(error_msg)



"""Package info"""

__all__ = [
    '__package__', '__command__', '__version__', '__min_python_version__',
    '__min_git_version__', '__url__',
]

#: Package name
__package__ = 'git-workflow'
#: Base command
__command__ = 'workflow'
#: Package version
__version__ = '0.2.1'
#: Minimum required python version
__min_python_version__ = '3.6'
#: Minimum required git version for advanced features
__min_git_version__ = 2.23
#: Project URL
__url__ = 'https://github.com/connordelacruz/git-workflow'



import os
from setuptools import setup, find_packages

base_path = os.path.dirname(os.path.abspath(__file__))
# Get metadata
about = {}
with open(os.path.join(base_path, 'git_workflow', '__about__.py')) as f:
    exec(f.read(), about)
# Parse README.rst for long_description
with open(os.path.join(base_path, 'README.rst')) as f:
    readme = f.read()

setup(
    name='git_workflow',
    version=about['__version__'],
    author='Connor de la Cruz',
    author_email='connor.c.delacruz@gmail.com',
    description='Utilities for streamlining our git workflow.',
    long_description=readme,
    url=about['__url__'],
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'Topic :: Software Development :: Version Control :: Git',
        'Topic :: Software Development',
        'Development Status :: 3 - Alpha',
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            about['__command__'] + ' = git_workflow.__main__:main',
        ]
    },
    install_requires=[
        'GitPython==3.1.11',
        'blessings>=1.7,<1.8',
        'argcomplete>=1.12,<1.13',
    ],
    extras_require={
        'dev': [
            'Jinja2>=2.11,<2.12',
            'build',
            'vermin',
        ],
    },
    python_requires='>=' + about['__min_python_version__'],
)




import os
from git_workflow.utils import cmd, files
from .base import WorkflowBase


class SetTemplate(WorkflowBase):
    """\
    Create and configure commit template for the current branch.

    By default, the commit template includes the specified ticket number before
    the commit message. E.g. for ticket number ``AB-12345``:

    ::

        [AB-12345] <commit message text starts here>

    The commit template file will be created in the root of the git repository.
    By default, the filename will be in this format:

    ::

        .gitmessage_local_<ticket>_<branch>

    The format of the filename, commit template body, accepted ticket numbers,
    and more can be customized with git configs (see the Configs section below
    for details).
    """
    command = 'set-template'
    description = 'Configure git commit template for a branch.'
    configs_used = [
        'commitTemplateFilenameFormat',
        'commitTemplateFormat',
        'ticketInputFormatRegex',
        'ticketFormatCapitalize',
        'ticketInputFormatRegex',
        'initials',
    ]

    @classmethod
    def add_subparser(cls, subparsers, generic_parent_parser):
        commit_template_subparser = cls._add_base_subparser(subparsers, generic_parent_parser)
        positional_args = commit_template_subparser.add_argument_group(
            'Positional Arguments'
        )
        positional_args.add_argument(
            'ticket', metavar='<ticket>', nargs='?', help='Ticket number to use in commit template',
            default=None
        )

    def get_args(self):
        """Parse command line arguments and prompt for any missing values.

        :return: A dictionary with the following keys:
            ticket
        """
        args = {}

        validate_ticket_number = cmd.generate_validate_regex_function(
            self.configs.TICKET_INPUT_FORMAT_REGEX
        )
        ticket = cmd.prompt(
            'Ticket Number',
            'Enter ticket number to use in commit messages.',
            invalid_msg='Invalid ticket number formatting.',
            initial_input=self.parsed_args.ticket,
            validate_function=validate_ticket_number,
            format_function=self.format_ticket_number,
        )
        args['ticket'] = ticket

        return args

    def run(self):
        args = self.get_args()
        repo_root_dir = os.path.dirname(self.repo.git_dir)
        branch_name = self.repo.active_branch.name
        # Create commit template
        format_kwargs = self.get_format_kwargs(args, branch_name)
        # NOTE: filenames will always begin with '.gitmessage_local_'
        commit_template_file = files.sanitize_filename(
            '.gitmessage_local_' + self.configs.COMMIT_TEMPLATE_FILENAME_FORMAT.format(**format_kwargs)
        )
        commit_template_path = os.path.join(repo_root_dir, commit_template_file)
        self.print('Creating commit template file...')
        commit_template_body = self.configs.COMMIT_TEMPLATE_FORMAT.format(**format_kwargs)
        with open(commit_template_path, 'w') as f:
            f.write(commit_template_body)
        if not os.path.exists(commit_template_path):
            raise Exception('Unable to create commit template at path ' + commit_template_path)
        self.print_success('Template file created.', commit_template_path, '')
        # Configure commit template
        # TODO REPHRASE OUTPUT. Current output would be fine for --verbose but is too much otherwise
        branch_config_file = files.sanitize_filename(f'config_{branch_name}')
        branch_config_path = os.path.join(self.repo.git_dir, branch_config_file)
        # Branch config
        self.print(f'Configuring commit.template for branch {branch_name}...')
        self.repo.git.config('commit.template', commit_template_file,
                             file=branch_config_path)
        self.print_success(f'commit.template configured in .git/{branch_config_file}.', '')
        # Add includeIf for branch config to local config
        self.print('Configuring local repo...')
        self.repo.git.config(f'includeIf.onbranch:{branch_name}.path', branch_config_file,
                             file=self.configs.CONFIG_PATH)
        self.print_success('Local repo configured.',
                           f'Will include branch config .git/{branch_config_file}',
                           f'when branch {branch_name} is checked out.',
                           '')

    # Helper Methods

    def format_ticket_number(self, val):
        """Format ticket number input based on configs.

        :param val: Text to format

        :return: formatted text
        """
        if self.configs.TICKET_FORMAT_CAPITALIZE:
            val = val.upper()
        return val

    def get_format_kwargs(self, args, branch_name):
        """Returns a dict mapping placeholders to their respective values.

        :param args: get_args() result
        :param branch_name: Name of the branch to create template for

        :return: Dictionary to pass as kwargs to .format()
        """
        format_kwargs = {
            'ticket': args['ticket'],
            'branch': branch_name,
            'initials': self.configs.INITIALS or '',
        }
        return format_kwargs
